# -*- coding: utf-8 -*-

import time
import intelhex
import sys



class Programmer:
    debug = False
    
    # be careful using these arrays, spi uses inline modification
    PROGRAMMING_ENABLE = [0xAC, 0x53, 0x00, 0x00]
    PROGRAMMING_CHIP_ERASE = [0xAC, 0x80, 0x00, 0x00]
    PROGRAMMING_POLL_RDY = [0xF0, 0x00, 0x00, 0x00]
    
    PROGRAMMING_READ_MEMORY_HIGH_3 = 0x28
    PROGRAMMING_READ_MEMORY_LOW_3 = 0x20
    
    PROGRAMMING_LOAD_PROGRAM_MEMORY_PAGE_HIGH = 0x48
    PROGRAMMING_LOAD_PROGRAM_MEMORY_PAGE_LOW = 0x40
     
    PROGRAMMING_WRITE_PROGRAM_MEMORY_PAGE = 0x4C
    
    PROGRAMMING_READ_LOCK_BITS = [0x58, 0x00, 0x00, 0x00]
    PROGRAMMING_READ_FUSE_BITS = [0x50, 0x00, 0x00, 0x00]
    PROGRAMMING_READ_FUSE_HIGH_BITS = [0x58, 0x08, 0x00, 0x00]
    PROGRAMMING_READ_EXTENDED_FUSE_BITS = [0x50, 0x08, 0x00, 0x00]
    PROGRAMMING_READ_CALIBRATION_BYTE = [0x38, 0x00, 0x00, 0x00]

    PROGRAMMING_WRITE_FUSE_BITS = [0xac, 0xa0, 0x00]
    
    # definitions for 328
    PROGRAMMING_PAGE_SIZE = 64 
    PROGRAMMING_PAGE_NUMBER = 256
    
    def __init__(self, peripherie):
        self.peripherie = peripherie

    def reset_on(self):
        self.peripherie.program_reset_on()

    def reset_off(self):
        self.peripherie.program_reset_off()
        
    def high(self, i):
        return (i >> 8) & 0xff
    
    def low(self, i):
        return i & 0xff

    def _decode(self, byte, pos, label, null, eins):
        if byte & (1 << pos) == 0:
            print("    {label: <20s} 0 {null:s}".format(label=label, null=null))
        else:
            print("    {label: <20s} 1 {eins:s}".format(label=label, eins=eins))
        
    def programming_readFuses(self):
        if self.debug:
            print("programming_readFuses")
        if self.programming_enable() == False:
            return
        r = self.peripherie.program_spi_xfer(self.PROGRAMMING_READ_CALIBRATION_BYTE)
        print("PROGRAMMING_READ_CALIBRATION_BYTE   {0:02x} {0:08b}".format(r[3]))
        
        r = self.peripherie.program_spi_xfer(self.PROGRAMMING_READ_EXTENDED_FUSE_BITS)
        print("PROGRAMMING_READ_EXTENDED_FUSE_BITS {0:02x} {0:08b}".format(r[3]))
        
        # Decoding ist CPU-spezifisch, hier 328
        self._decode(r[3], 0, "BODLEVEL0", "ENABLED", "")
        self._decode(r[3], 1, "BODLEVEL1", "ENABLED", "")
        self._decode(r[3], 1, "BODLEVEL2", "ENABLED", "")
             
        r = self.peripherie.program_spi_xfer(self.PROGRAMMING_READ_FUSE_BITS)
        print("PROGRAMMING_READ_FUSE_BITS          {0:02x} {0:08b}".format(r[3]))
        self._decode(r[3], 0, "CKSEL0", "ENABLED", "")
        self._decode(r[3], 1, "CKSEL1", "ENABLED", "")
        self._decode(r[3], 2, "CKSEL2", "ENABLED", "")
        self._decode(r[3], 3, "CKSEL3", "ENABLED", "")
        self._decode(r[3], 4, "SUT0  ", "ENABLED", "")
        self._decode(r[3], 5, "SUT1  ", "ENABLED", "")
        self._decode(r[3], 6, "CKOUT ", "ENABLED", "")
        self._decode(r[3], 7, "CKDIV8", "ENABLED", "")
        
        r = self.peripherie.program_spi_xfer(self.PROGRAMMING_READ_FUSE_HIGH_BITS)
        print("PROGRAMMING_READ_FUSE_HIGH_BITS     {0:02x} {0:08b}".format(r[3]))
        self._decode(r[3], 0, "BOOTRST", "ENABLED", "")
        self._decode(r[3], 1, "BOOTSZ0", "ENABLED", "")
        self._decode(r[3], 2, "BOOTSZ1", "ENABLED", "")
        self._decode(r[3], 3, "EESAVE ", "ENABLED", "")
        self._decode(r[3], 4, "WDTON  ", "ENABLED", "")
        self._decode(r[3], 5, "SPIEN  ", "ENABLED", "")
        self._decode(r[3], 6, "DWEN   ", "ENABLED", "")
        self._decode(r[3], 7, "RSTDISBL", "ENABLED", "")

        r = self.peripherie.program_spi_xfer(self.PROGRAMMING_READ_LOCK_BITS)
        print("PROGRAMMING_READ_LOCK_BITS          {0:02x} {0:08b}".format(r[3]))

        self.programming_disable()
        if self.debug:
            print("programming_readCode ende")

    def programming_writeFuse(self, value):
        if self.debug:
            print("programming_writeFuse")
        if self.programming_enable() == False:
            return

        self._decode(value, 0, "CKSEL0", "ENABLED", "")
        self._decode(value, 1, "CKSEL1", "ENABLED", "")
        self._decode(value, 2, "CKSEL2", "ENABLED", "")
        self._decode(value, 3, "CKSEL3", "ENABLED", "")
        self._decode(value, 4, "SUT0  ", "ENABLED", "")
        self._decode(value, 5, "SUT1  ", "ENABLED", "")
        self._decode(value, 6, "CKOUT ", "ENABLED", "")
        self._decode(value, 7, "CKDIV8", "ENABLED", "")

        cmd = list(self.PROGRAMMING_WRITE_FUSE_BITS)
        cmd.append(value)
        r = self.peripherie.program_spi_xfer(cmd)

        self.programming_disable()
        if self.debug:
            print("programming_readCode ende")
 
         
    def programming_readCode(self, fileName):
        if self.debug:
            print("programming_readCode")
            
        if self.programming_enable() == False:
            return
        
        ihex = intelhex.IntelHex()
        
        for i in range(0x8000 / 2):
            r = self.peripherie.program_spi_xfer([self.PROGRAMMING_READ_MEMORY_HIGH_3, self.high(i), self.low (i), 0])
            h = r[3]
            r = self.peripherie.program_spi_xfer([self.PROGRAMMING_READ_MEMORY_LOW_3, self.high(i), self.low (i), 0])
            l = r[3]
            
            ihex[i * 2 + 0] = l
            ihex[i * 2 + 1] = h
        ihex.write_hex_file(fileName)
            
        self.programming_disable()
        if self.debug:
            print("programming_readCode ende")
    
    def programming_writeCode(self, fileName):
        if self.debug:
            print("programming_writeCode", fileName)
            
        if self.programming_enable() == False:
            return
        
        ihex = intelhex.IntelHex()
        ihex.loadhex(fileName)
        
        for page in range(self.PROGRAMMING_PAGE_NUMBER):
            # load page
            for index in range(self.PROGRAMMING_PAGE_SIZE):
                imem = (self.PROGRAMMING_PAGE_SIZE * page + index) * 2
                
                l = ihex[imem + 0]
                h = ihex[imem + 1]
                # if self.debug:
                #     print("{addr:04x} {low:02x} {high:02x}".format(addr=imem,low=l,high=h))
                
                r = self.peripherie.program_spi_xfer(
                                  [self.PROGRAMMING_LOAD_PROGRAM_MEMORY_PAGE_LOW,
                                   0x00,
                                   index,
                                   l])
                r = self.peripherie.program_spi_xfer(
                                  [self.PROGRAMMING_LOAD_PROGRAM_MEMORY_PAGE_HIGH,
                                   0x00,
                                   index,
                                   h])
            
            if True:
                # write page
                n = page * self.PROGRAMMING_PAGE_SIZE
                r = self.peripherie.program_spi_xfer(
                                  [self.PROGRAMMING_WRITE_PROGRAM_MEMORY_PAGE,
                                   self.high(n),
                                   self.low(n),
                                   0])
                self.programming_poll()
        if self.debug:
            print("programming_writeCode write")
        
    def programming_disable(self):
        if self.debug:
            print("programming_disable")
        self.reset_off()
        time.sleep(0.02)

    def programming_enable(self):
        if self.debug:
            print("programming_enable")
        
        LOOP = 4
        for loop in range (LOOP):
            # apply reset
            self.reset_on()
          
            # wait at least 20 ms
            time.sleep(0.021)
            print("PROGRAMMING_ENABLE", self.PROGRAMMING_ENABLE)
            r = self.peripherie.program_spi_xfer(list(self.PROGRAMMING_ENABLE))
            
            if self.debug:
                print(loop, r)
                
            if r[2] == 0x53:
                # in sync
                if self.debug:
                    print("programming_enable end")
                return True
            else:
                self.reset_off()
                time.sleep(0.1)
                
        if self.debug:
            print ("not in sync, tried {0:d} times".format(LOOP))
        self.reset_off()
        if self.debug:
            print("programming_enable end")
        return False
            
    def programming_clearCode(self):
        if self.debug:
            print("programming_clearCode")
        if self.programming_enable() == False:
            return
   
        r = self.peripherie.program_spi_xfer(list(self.PROGRAMMING_CHIP_ERASE))
        self.programming_poll()
        
        self.programming_disable()
        if self.debug:
            print("programming_clearCode end")
        
    def programming_poll(self):
        # if self.debug:
        #    print("programming_poll")

        # es werden max 9 ms Zeitverzoegerung ben�tigt.
        
        for i in range(11):
            r = self.peripherie.program_spi_xfer(list(self.PROGRAMMING_POLL_RDY))
            # if self.debug:
            #    print('poll_rdy', r)
            if r[3] & 1 == 0:
                break
            time.sleep(0.001) 
